#ifndef AK_INPUT_STREAM_HPP
#define AK_INPUT_STREAM_HPP

#include <iostream>
#include "ref_ptr.hpp"

using std::istream;
using std::streambuf;

/**
  * A <i>InputStream<i> object is a <b>reference-count</b> pointer that wraps
  * a <i>istream</i>. Provides services indirectly from istream.  
  * 
  * Clients should not assume anything about the type of the 
  * stream wrapped in this object.
  * 
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class InputStream : public ref_ptr<istream> {

public:

  /**
    * Constructs this object. 
    *
    * @param impl the istream for input
    */
  InputStream(istream* impl = NULL) 
    : ref_ptr<istream> (impl) {
  }

  /**
    * Decrements the ref_cnt by 1, and when it goes to zero destroys the
    * istream and this object.
    */
  ~InputStream() {
  }

  /**
    * The overloaded input operator, extracts a c-string from the input stream. 
    *
    * @param c the c-string 
    * @return the input stream
    */
  InputStream& operator>>(char* c) {
    *get() >> c;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a character from the input
    * stream. 
    * 
    * @param c the character 
    * @return the input stream
    */
  InputStream& operator>>(char& c) {
    *get() >> c;
    return *this;
  }

  /**
    * The overloaded input operator, extracts an integer from the input stream.
    *
    * @param i the integer
    * @return the input stream
    */
  InputStream& operator>>(int& i) {
    *get() >> i;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a long integer from the 
    * input stream.
    *
    * @param l the long integer 
    * @return the input stream
    *
    */
  InputStream& operator>>(long& l) {
    *get() >> l;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a short integer from the 
    * input stream.
    *
    * @param short& s the short integer
    * @return InputStream& the input stream
    */
  InputStream& operator>>(short& s) {
    *get() >> s;
    return *this;
  }

  /**
    * The overloaded input operator, extracts an unsigned integer from the 
    * input stream.
    *
    * @param ui the unsigned integer
    * @return the input stream
    */
  InputStream& operator>>(unsigned int& ui) {
    *get() >> ui;
    return *this;
  }

  /**
    * The overloaded input operator, extracts an unsigned long integer from the
    * input stream.
    *
    * @param ui the unsigned long integer
    * @return the input stream
    */
  InputStream& operator>>(unsigned long& ul) {
    *get() >> ul;
    return *this;
  }

  /**
    * The overloaded input operator, extracts an unsigned short integer from the
    * input stream.
    *
    * @param us the unsigned short integer
    * @return the input stream
    */
  InputStream& operator>>(unsigned short& us) {
    *get() >> us;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a float from the
    * input stream.
    *
    * @param f the float 
    * @return the input stream
    */
  InputStream& operator>>(float& f) {
    *get() >> f;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a double from the input stream.
    *
    * @param d the double 
    * @return the input stream
    */
  InputStream& operator>>(double& d) {
    *get() >> d;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a long double from the 
    * input stream.
    *
    * @param ld the long double 
    * @return the input stream
    */
  InputStream& operator>>(long double& ld) {
    *get() >> ld;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a boolean from the input stream.
    *
    * @param b the boolean 
    * @return the input stream
    */
  InputStream& operator>>(bool& b) {
    *get() >> b;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a stream buffer from the
    * input stream.
    *
    * @param sb the stream buffer 
    * @return the input stream
    */
  InputStream& operator>>(streambuf* sb) {
    *get() >> sb;
    return *this;
  }

  /**
    * The overloaded input operator, extracts a string from the input stream.
    *
    * @param s the string  
    * @return the input stream
    */
  InputStream& operator>>(string& s) {
    *get() >> s;
    return *this;
  }
  
  /**
    *
    */
  string getline(char delim = '\n') {
    string line;

    std::getline(*get(), line, delim);
    return line;
  }

};

} // end of namespace Akenti

#endif
