#ifndef AK_OUTPUT_STREAM_HPP
#define AK_OUTPUT_STREAM_HPP

#include <iostream>
#include "ref_ptr.hpp"

using std::ostream;

namespace Akenti {

/**
  * A <i>OutputStream</i> object is a <i>reference-count</i>
  * pointer that wraps a <i>ostream</i> object.
  * 
  * Clients should not assume anything about the type of the 
  * stream wrapped in this object.
  *
  * @version 1.1 00/11/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  * 
  * @see ref_ptr
  */

class OutputStream : public ref_ptr<ostream> {

public:

  /**
    * Constructs this object. 
    * 
    * @param impl the ostream for output 
    */ 
  OutputStream(ostream* impl = NULL)
    : ref_ptr<ostream> (impl) {
  }

  /**
    * Decrements the ref_cnt by 1, when it drops to 0 destroys the ostream
    * and this object.
    */
  ~OutputStream() {
  }

  /**
    * The overloaded output operator, that places a character into the 
    * output stream.
    *
    * @param c the character
    * @return the output stream
    */
  OutputStream& operator<<(const char c) {
    *get() << c;
    return *this;
  }

  /**
    * The overloaded output operator, that places a C-string into the 
    * output stream.
    *
    * @param c the C-string 
    * @return the output stream
    */
  OutputStream& operator<<(const char* s) {
    *get() << s;
    return *this;
  }

  /**
    * The overloaded output operator, that places the first thing that the 
    * pointer points to into the output stream.
    *
    * @param p the void * pointer 
    * @return the output stream
    */
  OutputStream& operator<<(const void *p) {
    *get() << p;
    return *this;
  }

  /**
    * The overloaded output operator, that places a character into the 
    * output stream.
    *
    * @param c the character
    * @return the output stream
    */
  OutputStream& operator<<(int i) {
    *get() << i;
    return *this;
  }

  /**
    * The overloaded output operator, that places an unsigned integer into the 
    * output stream.
    *
    * @param ui the unsigned integer 
    * @return the output stream
    */
  OutputStream& operator<<(unsigned int ui) {
    *get() << ui;
    return *this;
  }

  /**
    * The overloaded output operator, that places a long integer into the 
    * output stream.
    *
    * @param l the long integer 
    * @return the output stream
    */
  OutputStream& operator<<(long l) {
    *get() << l;
    return *this;
  }

  /**
    * The overloaded output operator, that places an unsigned long integer 
    * into the output stream.
    *
    * @param ul the unsigned long integer 
    * @return the output stream
    */
  OutputStream& operator<<(unsigned long ul) {
    *get() << ul;
    return *this;
  }

  /**
    * The overloaded output operator, that places a boolean into the 
    * output stream.
    *
    * @param b the boolean 
    * @return the output stream
    */
  OutputStream& operator<<(bool b) {
    *get() << b;
    return *this;
  }
  
  /**
    * The overloaded output operator, that a double into the output stream.
    *
    * @param d the double 
    * @return the output stream
    */
  OutputStream& operator<<(double d) {
    *get() << d;
    return *this;
  }
  
  /**
    * The overloaded output operator, that the stream buffer into the 
    * output stream.
    *
    * @param sb the stream buffer 
    * @return the output stream
    */
  OutputStream& operator<<(std::streambuf* sb) {
    *get() << sb;
    return *this;
  }

  /**
    * The overloaded output operator, that places a string into the 
    * output stream.
    *
    * @param s the string  
    * @return the output stream
    */
  OutputStream& operator<<(const string& s) {
    *get() << s;
    return *this;
  }

};

} // end of namespace Akenti

#endif
