#ifndef AK_SSL_CONTEXT_HPP
#define AK_SSL_CONTEXT_HPP

#include "ref_ptr.hpp"
#include "Error.hpp"
#include "SSLContextImpl.hpp"

namespace Akenti {

/**
  * A ref_ptr class to wrap a security context. A security context contains the
  * the values that a secure tranport protocol such as TLS needs. It may include
  * such things as the private key and identity certificate of the principal 
  * represented, what level of peer certification is required, the 
  * required levels of message security, the ciphers supported, and the 
  * CA certificates of trusted CA's.
  * The actual context is implmented in a concrete SSLcontext which is a 
  * subclass of the SSLContextImpl.
  *
  * @see ref_ptr
  * @see SSLContextImpl
  * 
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class SSLContext : public ref_ptr<SSLContextImpl> {

public:

  /**
    * Constructor 
    * 
    * @param a reference to a SSLContextImpl
    */
  SSLContext(SSLContextImpl* impl = NULL) 
    : ref_ptr<SSLContextImpl> (impl) {
  }

  /**
    * Destroys this object.
    */ 
  ~SSLContext() {
  }

  /**
    * Sets a default server security context.
    */
  static void
  setDefaultSSLServerContext(const SSLContext& sc) {
    if (!sc || sc->isClientContext()) {
      throw Error("ProviderImpl:setDefaultSSLContext:invalid arg");
    }

    SSLContext::serverContext = sc;
  }

  /**
    * Returns the default server security context.
    */
  static const SSLContext&
  getDefaultSSLServerContext() {
    if (!serverContext) {
      throw Error("ProviderImpl:getDefaultSSLServerContext:context not set");
    }

    return serverContext;
  }

  /**
    * Sets a default client security context.
    */
  static void
  setDefaultSSLClientContext(const SSLContext& cc) {
    if (!cc || !cc->isClientContext()) {
      throw Error("ProviderImpl:setDefaultSSLClientContext:invalid arg");
    }

    SSLContext::clientContext = cc;
  }

  /**
    * Returns the default server security context.
    */
  static const SSLContext&
  getDefaultSSLClientContext() {
    if (!clientContext) {
      throw Error("ProviderImpl:getDefaultSSLClientContext:context not set");
    }

    return clientContext;
  }

private:
 
  static SSLContext serverContext;

  static SSLContext clientContext;

};

} // end of namespace Akenti

#endif
