#ifndef AKENTI_SERVER_CONTEXT_HPP
#define AKENTI_SERVER_CONTEXT_HPP

#include <string>

#include "PrivateKey.hpp"
#include "Certificate.hpp"

using std::string;
using namespace Akenti;

namespace Akenti {

/**
  * Class to hold the Akenti server credentials which are used to 
  * sign cache and capability credentials. One such object is statically
  * created. 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */
class ServerContext {

public:
  
  /**
    * destructor
    */
  ~ServerContext();

  /**
    * @return the Akenti server's private key
    * @exception Error if the private key is not loaded
    */
  const PrivateKey& getPrivateKey() const; 

  /**
    * @return the Akenti server's x509 id certificate
    * @exception Error if the certificate is not loaded
    */
  const Certificate& getCertificate() const;

  /**
    * @return true if both the private key and certificate have been loaded
    *          false otherwise
    */
  bool areCredentialsLoaded() const;

  /**
    * Reads the file given by fileName and initializes the private key
    *
    * @param fileType - currently only "PEM" is supported which means
    *      the PEM tagged, base 64-encoded DER representation of a private
    *      key. The private keys are interpreted by openSSL which accepts
    *      unencrypted RSA, DSA or PKCS8 keys.
    * @param fileName - absolute pathname of the file containing the 
    *       private key
    * @exception Error if the private key has already been set
    */
  void 
  initPrivateKeyFromFile(const string& fileType, 
                         const string& fileName);
  /**
    * Reads in a certificate from a file
    * 
    * @param fileType the certificate encoding to accept - currently we
    *        only support "PEM" which is the standard PEM tagged base64
    *        encoding of the DER encoding of the ANS1 x509 certificate
    * @param fileName - absolute pathname of the file containing the 
    *       certificate
    * @exception Error if the certificate has already been set
    */
  void 
  initCertificateFromFile(const string& fileType, 
                          const string& fileName);
  /**
    * static method to return the single ServerContext that is statically
    * created.
    */
  static ServerContext& getServerContext();

  /**
   * @return true if the credentials have been loaded, false otherwise
    */
  static bool isServerContextSet();

private:

  friend class AkentiInit;

  /**
    */
  ServerContext();

  /**
    */
  ServerContext(const ServerContext&);

  /**
    */
  void operator=(const ServerContext&);

  /**
    */
  void unloadCredentials();

  PrivateKey privateKey;

  Certificate certificate;

  static ServerContext context;
};

} // end of name space

#endif

