#include <iostream> 
#include <sstream> 
#include <fstream> 
#include <unistd.h>
#include "AkentiCertificate.hpp"
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "KeyStore.hpp"
#include "File.hpp"
#include "Certificate.hpp"
#include "AttributeCertificate.hpp"
#include "AttributeCertGenerator.hpp"
#include "PolicyCertificate.hpp"
#include "PolicyCertGenerator.hpp"
#include "UseCondCertificate.hpp"
#include "UseCondCertGenerator.hpp"
#include "Expression.hpp"
#include "XMLUtil.hpp"
#include "AkentiConstants.hpp"
#include "Util.hpp"
#include "Exception.hpp"

using namespace Akenti;
using namespace std;

/**
  * <pre>
  * <b>CertCen</b>
  * Program to generate and sign a certificate from an XML input file.
  *
  * Note: The xml input file you provide must contain a valid xml 
  *       encoded certificate according to the akenti certificate schema.
  *       Otherwise the certificate parser would refuse it.
  *       Edit existing certificates to avoid problems. And don't forget to
  *       leave the old signature element intact.
  *
  * Usage: CertGen -k keyfile -p passwordFile -i xmlInputFile [-o outfile] [-d days] [-h] [-v]
  * </pre>
  */

namespace CertGen {

string computeUniqueID(const AkentiCertificate& cert) {
  string hash = cert->hashCode();

  return Util::createUniqueID(Util::createFileName(hash));
}

pair<UtcTime, UtcTime> 
computeValidityTimes(const AkentiCertificate& cert, int period) {
    UtcTime now;

    if (period <= 0) {
      UtcTime notBefore = cert->getNotBefore();

      if (notBefore < now) {
        notBefore = now;
      }

      UtcTime notAfter = cert->getNotAfter();

      if (notBefore >= notAfter) {
        throw Exception("Bad expiration time");
      }

      return std::make_pair(notBefore, notAfter);
    }

    return make_pair(now, now + period);
}

int toInt(const string& s) {
  istringstream is(s);
  int ret;

  if (!(is >> ret)) {
    throw Exception("Excpecting a number");
  }
  
  return ret;
}

AkentiCertificate 
generateAttributeCertificate(const AttributeCertificate& cert, const KeyStore& ks, int period) {
  AttributeCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID(computeUniqueID(cert));

  const Certificate& issuerCert = ks->getSubjectCertificate();
  AkentiPrincipal issuer(issuerCert->getPrincipal(), issuerCert->getGuarantor());
  
  gen.setIssuer(issuer);
  
  pair<UtcTime, UtcTime> validityTimes = computeValidityTimes(cert, period);

  gen.setNotBefore(validityTimes.first);
  gen.setNotAfter(validityTimes.second);

  gen.setPrincipal(cert->getPrincipal());
  gen.setAttribute(cert->getAttribute());
  gen.setValue(cert->getValue());
  
  if (cert->containsCondition()) {
    gen.setCondition(cert->getCondition());
  }
  
  PrivateKey privateKey = ks->getPrivateKey();

  gen.sign(privateKey, "RSA-MD5");

  return gen.generate();
}

AkentiCertificate 
generatePolicyCertificate(const PolicyCertificate& cert, const KeyStore& ks, int period) {
  PolicyCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID(computeUniqueID(cert));

  const Certificate& issuerCert = ks->getSubjectCertificate();
  AkentiPrincipal issuer(issuerCert->getPrincipal(), issuerCert->getGuarantor());
  
  gen.setIssuer(issuer);
  
  pair<UtcTime, UtcTime> validityTimes = computeValidityTimes(cert, period);

  gen.setNotBefore(validityTimes.first);
  gen.setNotAfter(validityTimes.second);
  gen.setResource(cert->getResource());

  const vector<CertificateAuthorityInfo>& infos = cert->getCertificateAuthorityInfos();

  if (infos.size()) {
    gen.setCertificateAuthorityInfos(infos);
  }

  gen.setUseCondIssuerGroups(cert->getUseCondIssuerGroups());

  const vector<Directory>& attrDirs = cert->getAttributeDirectories();

  if (attrDirs.size()) {
    gen.setAttributeDirectories(attrDirs);
  }

  gen.setCacheTime(cert->getCacheTime());

  PrivateKey privateKey = ks->getPrivateKey();

  gen.sign(privateKey, "RSA-MD5");
  return gen.generate();
}

AkentiCertificate 
generateUseCondCertificate(const UseCondCertificate& cert, const KeyStore& ks, int period) {
  UseCondCertGenerator gen;

  gen.setVersionNumber(2);
  gen.setUniqueID(computeUniqueID(cert));

  const Certificate& issuerCert = ks->getSubjectCertificate();
  AkentiPrincipal issuer(issuerCert->getPrincipal(), issuerCert->getGuarantor());
  
  gen.setIssuer(issuer);
  
  pair<UtcTime, UtcTime> validityTimes = computeValidityTimes(cert, period);

  gen.setNotBefore(validityTimes.first);
  gen.setNotAfter(validityTimes.second);
  gen.setResource(cert->getResource());
  gen.setCritical(cert->isCritical());
  gen.setLocalScope(cert->isScopeLocal());

  const Condition& cond = cert->getCondition();
  Expression expr;

  if (!cond.parseConstraint(expr)) {
    throw Exception("Invalid boolean expression");
  }

  if (!cond.isValid()) {
    throw Exception("Invalid condition/attributes infos");
  }

  gen.setCondition(Condition(expr.toString(), cond.getAttributeInfos()));
  gen.setActions(cert->getActions());

  PrivateKey privateKey = ks->getPrivateKey();

  gen.sign(privateKey, "RSA-MD5");
  return gen.generate();
}

void printPurpose() {
  cerr << "Purpose: Use this program to generate and sign Policy," << endl
       << "         Use-Condition, and Attribute certs." << endl
       << endl
       << "Note: The xml input file you provide must contain " << endl
       << "      a valid xml-encoded akenti certificate." << endl
       << "      Use existing certificates/templates" << endl
       << endl;
}

void printUsage () {
  cerr << "Usage: CertGen args" << endl
       << "  -k arg        - keyfile in p12 format" << endl
       << "  -p arg        - file where passphrase is stored" << endl 
       << "  -i arg        - xml input file" << endl 
       << "  [-o arg]      - friendly name to store file" << endl
       << "  [-d arg]      - number of days" << endl 
       << "  [-v]          - verbose" << endl
       << "  [-h]          - prints this message" << endl; 
}

int main(int argc, char* argv[]) {
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  string keyfile;
  string passwordFile; 
  string outfile;
  string xmlfile;
  int period = 0; 
  bool verbose = false;
  int opt;

  while ((opt = getopt(argc, argv, "k:p:o:i:d:vh")) != EOF) {
    switch(opt){
      case 'k':
        keyfile = optarg;
        break;
      case 'p':
        passwordFile = optarg;
        break;
      case 'o':
        outfile = optarg;
        break;
      case 'i':
        xmlfile = optarg;
        break;
      case 'd':
        period += toInt(optarg) * 24 * 60 * 60;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
    }
  }

  try {
    if (!keyfile.length()) {
      cerr << "Must provide a p12 file" << endl;
      exit(2);
    } else if (!passwordFile.length()) {
      cerr << "Must provide a passwordFile" << endl;
      exit(2);
    }

    if (verbose) { 
      cout << "Initializing xml lib..." << endl;
    }

    XMLUtil::init();

    if (verbose) { 
      cout << "Setting security provider ..." << endl;
    }

    Provider provider = Provider(new OpenSSLProviderImpl);

    provider->loadDigestAlgorithms();
    Provider::setProvider(provider);

    if (verbose) { 
      cout << "Loading p12file ..." << File(keyfile).getAbsolutePath() << endl;
    }

    KeyStore ks = KeyStore::getInstance("PKCS12");

    ifstream is(passwordFile.c_str());
    string temp;

    // is >> temp;
    std::getline(is, temp);

    if (verbose) { 
      cout << "Loading p12file ..." << File(keyfile).getAbsolutePath() << endl;
    }

    ks->load(keyfile, temp);

    if (!xmlfile.length()) {
      cerr << "Must provide an xml input file" << endl;
      exit(2);
    }

    if (verbose) {
      cout << "Parsing XML input file" + File(xmlfile).getAbsolutePath() << endl;
    }

    AkentiCertificate inCert = AkentiCertificate::initFromFile("XML", xmlfile);
    AkentiCertificate outCert;

    if (inCert.instanceOf<AttributeCertificate>()) {
      outCert = generateAttributeCertificate((AttributeCertificate&) inCert, ks, period);
    } else if (inCert.instanceOf<UseCondCertificate>()) {
      outCert = generateUseCondCertificate((UseCondCertificate&) inCert, ks, period);
    } else if (inCert.instanceOf<PolicyCertificate>()) {
      outCert = generatePolicyCertificate((PolicyCertificate&) inCert, ks, period);
    } else {
      throw Exception("Can't handle cert type " + inCert->getType());
    }

    if (!inCert.instanceOf<PolicyCertificate>()) {
      if (outfile.length() != 0) {
        outfile += '.';
      }

      outfile += Util::createFileName(outCert->hashCode());
      outfile += '.';

      int i = 0;

      while (true) {
        ostringstream os;
      
        os << i++; 

        if (!File(outfile + os.str()).exists()) {
          outfile += os.str();
          break;
        }
      }
    } else if (!outfile.length()) {
        outfile = PolicyFileName;
    }

    cout << "Writing cert to " << File(outfile).getAbsolutePath() << endl;

    ofstream os(outfile.c_str());

    AkentiCertificate::writeToStream("XML", os, outCert);
    os.close();
    exit(0);
  } catch(Exception& ex) {
    cerr << ex << endl;
    exit(2);
  } catch (Error& error) {
    cerr << "ERROR:Report to Akenti team:" <<  error << endl;
    exit(3);
  }
}

} 
