#include <sstream>
#include <iostream>
#include <fstream>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "CertificateChain.hpp"
#include "CheckAccessRequest.hpp"
#include "CheckAccessResponse.hpp"
#include "AkentiCertificate.hpp"
#include "CapabilityCertificate.hpp"
#include "ProtocolHandler.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "SSLContext.hpp"
#include "XMLUtil.hpp"

using namespace Akenti;
using namespace std;

/**
  * <pre>
  * <b>RemoteCheckAccess</b>
  *   Use this program to check a user's access to a remote resource.
  * </pre>
  */
namespace RemoteCheckAccess {

void
printActions(const vector<string>& actions, ostream *os) {
  *os << "Allowed actions:";

  int size = actions.size();

  if (size == 0) {
    cout << "NONE" << endl;
    return;
  }

  *os << actions[0];

  for (int i = 1; i < actions.size(); ++i) {
    *os << ", " << actions[i];
  }

  *os << endl;
}


void
printActions(const vector<ConditionalActions>& actions, ostream *os) {
  *os << "Conditional actions are: ";

  int size = actions.size();

  if (size == 0) {
    *os << "NONE" << endl;
    return;
  }

  *os << actions[0].toLogString();

  for (int i = 1; i < actions.size(); ++i) {
    //cerr << ", " << actions[i].toLogString();
    *os << ", " << actions[i].toLogString();
  }

  *os << endl;
}



void printPurpose() {
  cerr << "Use this program to check a user's access to a remote resource." << endl 
       << endl;
}

void printUsage() {
  cerr << "Usage: RemoteCheckAccess args" << endl
       << "  -F arg             - akenti engine config file" << endl
       << "  -r arg             - resource " << endl
       << "  [-x arg]           - user's x509 cert chain" << endl
       << "  [-u arg -c arg]    - user's DN and user's CADN" << endl 
       << "  [-R arg]*          - a specific action(s) " << endl
       << "  [-A arg -V arg]*   - attribute/value pair" << endl
       << "  [-C]               - show capability cert only" << endl
       << "  [-P]               - show all certs used in making the decision" << endl
       << "  [-o arg]           - certs' output file (defaults to stdout)" << endl
       << "  [-q]               - don't show cap in pretty text (the default)" << endl
       << "  [-v]               - verbose" << endl
       << "  [-h]               - prints this message" << endl;
}

int 
main(int argc, char** argv) {
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  string confFileName;
  string x509CertFileName;
  string userDN;
  string caDN;
  string resource;
  string certOutputFile;
  set<string> desiredActions;
  multimap<string, string> desiredAttributes;
  string attribute;
  string value;
  bool detailed = false;
  bool capOnly = false;
  bool verbose = false;
  bool concise = false;
  int opt;

  while ((opt = getopt(argc, argv, "r:x:u:c:F:o:A:V:R:CPvqh")) != EOF) {
    switch(opt) {
      case 'u':
        userDN = optarg;
        break;
      case 'c':
        caDN = optarg;
      case 'x':
        x509CertFileName = optarg;
        break;
      case 'F':
        confFileName = optarg;
        break;
      case 'r':
        resource = optarg;
        break;
      case 'o':
        certOutputFile = optarg;
        break;
      case 'R':
        desiredActions.insert(optarg);
        break;
      case 'A':
        attribute = optarg;
        break;
      case 'V':
        value = optarg;

        if (attribute.length() == 0) {
          cerr << "Must provide attribute and a value pair" << endl;
          exit(2);
        }

        desiredAttributes.insert(make_pair(attribute, value));
        attribute = value = "";
        break;
      case 'C':
        capOnly = true;
        break;
      case 'P':
        detailed = true;
        break;
      case 'q':
        concise = true;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
    }
  }

  if (attribute.length() != 0) {
    cerr << "Must provide attribute and a value pair" << endl;
    exit(2);
  }


  if ((!userDN.length() || !caDN.length()) && (!x509CertFileName.length())) {
    cerr << "Must provide either a userDN and userCaDN or x509CertFileName" << endl;
    exit(2);
  } else if (!resource.length()) {
    cerr << "Must provide a resource" << endl;
    exit(2);
  }

  try {

    if (verbose) {
     cout  << "Initiizing XML library ... " << endl;
    }
 

    XMLUtil::init();
 
    if (verbose) {
      cout  << "Installing security provider ..." << endl;
    }
 
    Provider po = Provider(new OpenSSLProviderImpl);
 
    po->loadSignatureAlgorithms();
    po->seedRandomGenerator("WE GOT TO SEED THE RANDOM GENERATOR");
    Provider::setProvider(po);
 
    if (verbose) {
      cout  << "Loading client configuration file ..." << endl;
    }

    Properties& props = Properties::getSystemProperties();
 
    props.load(confFileName);
 
    if (StringUtil::equalsIgnoreCase(props.getValue("SSLMode"), "on")) {
      if (verbose) {
        cout  << "Setting ssl context ..." << endl;
      }
 
      SSLContext ctx = po->createSSLContext(SSLConstants::TLS1, true);
 
      ctx->setVerify(SSLConstants::VERIFY_PEER);
      ctx->setCertificateFile(props.getValue("ClientCertFile"), SSLConstants::PEM);
      ctx->setPrivateKeyFile(props.getValue("ClientKeyFile"), SSLConstants::PEM);
 
      string path = props.getValue("CertificateAuthorityPath");
 
      if (path.length() != 0) {
        path += "/";
      }
 
      path += props.getValue("CertificateAuthorityFile");
      ctx->setTrustedCAS(path);
      SSLContext::setDefaultSSLClientContext(ctx);
    }

    /* First set either the Akenti Principal or a certificate chain 
     * - Cert Chain NOT IMPLEMENTED YET - so go ahead and set Principal
     * for now
     */

    AkentiPrincipal user;

    if (x509CertFileName != "") {
      if (verbose) {
        cout << "Loading user's cert chain from " << x509CertFileName << endl;
      }

      CertificateFactory certFac = CertificateFactory::getInstance();
      CertificateChain chain = certFac->initCertificateChainFromFile(
        "PEM", x509CertFileName
      );

      Certificate cert = chain->getSubjectCertificate();

      /* Need to set Akenti chain here in the request somehow */
      /* Remove next line when this is done */
      /* THIS PART NOT IMPLEMENTED YET */
      user = AkentiPrincipal(cert->getPrincipal(), cert->getGuarantor());
    } else {
       user = AkentiPrincipal(userDN, caDN);

       if (verbose) {
          cout << "User DN is " << userDN << " CaDN is "<< caDN << endl;
          cout << "Calling eng.process for resource " << resource << endl;
        }
     }


    CheckAccessRequest req;
    CheckAccessResponse resp;

    req.setAkentiPrincipal (user);
    req.setID(1);
    req.setResource(resource);
    req.setAkentiPrincipal(user);
    req.setDetailed (detailed);
    req.setDesiredActions (desiredActions);
    req.setRestrictedAttributes (desiredAttributes);

    ProtocolHandler::getInstance(ProtocolHandler::CLIENT).handle(req, resp);

    AkentiMessage mesg = resp.getAkentiMessage();
    CapabilityCertificate capCert = resp.getCapabilityCertificate();

    // set stream to output file if specified
    // cast needed for gcc 2.9.5 on freebsd 4.x
    bool haveOutputFile = certOutputFile.length() != 0;
    ostream* os = haveOutputFile? (ostream*) new ofstream(certOutputFile.c_str()) : &cout;

    if (!detailed) {
       if (capCert && concise) {
           printActions(capCert->getUnConditionalActions(), os);
           printActions(capCert->getConditionalActions(), os);
       } else if (capCert) {
           AkentiCertificate::writeToStream("XML", *os, capCert);
       }

       if (haveOutputFile) {
           *os << resp.getAkentiMessage().toString() << endl;
       }
       cout << resp.getAkentiMessage().toString() << endl;
    } else { // looking for details

      // write out Policy and associated use condition certificates
      const vector <PolicyInfo>& policyInfoVec = resp.getPolicyInfo();
      for (int i = 0; i < policyInfoVec.size(); ++i) {
          const vector <CertificateInfo>& certInfoVec = policyInfoVec[i].getCertificateInfos();
          for (int j = 0; j < certInfoVec.size(); ++j) {
              if (certInfoVec[j].hasCertificate()) {
                  AkentiCertificate::writeToStream("XML", *os, certInfoVec[j].getCertificate());
              }
          }
      }
  
      // At this point these are all attribute certs
      const vector<AkentiCertificate>& akentiCerts = resp.getAkentiCerts();

      for (int k = 0; k < akentiCerts.size(); ++k) {
        AkentiCertificate::writeToStream("XML", *os, akentiCerts[k]);
      }

      if (capCert) {
        AkentiCertificate::writeToStream("XML", *os, capCert);
      }

      if (haveOutputFile) {
        *os << mesg.toString() << endl;
        delete(os);
      }
      
      cout << mesg.toString() << endl;

    }

    exit(0);
  } catch(Exception& ex) {
    cerr << "Exception:" << ex << endl;
    exit(2);
  } catch(Error& err) {
    cerr << "ERROR:(Report to akenti team)" << err << endl;
    exit(3);
  } 
}

}
