#include <sstream>
#include <iostream>
#include <string>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "SSLContext.hpp"
#include "ProtocolHandler.hpp"
#include "AkentiCertificate.hpp"
#include "CheckPolicyRequest.hpp"
#include "CheckPolicyResponse.hpp"
#include "Properties.hpp"
#include "StringUtil.hpp"
#include "XMLUtil.hpp"

using namespace Akenti;
using namespace std;

/**
  * <pre>
  * <b>RemoteCheckPolicy:</b>
  *    Use this program to remotely check resource policy.
  * </pre>
  */

namespace RemoteCheckPolicy {

void printPurpose() {
  cerr << "Use this program to remotely collect resource policy." << endl 
       << endl;
}

void printUsage () {
  cerr << "Usage: RemoteCheckPolicy args" << endl
       << "  -F arg   - client configuration file" << endl
       << "  -r arg   - resource" << endl
       << "  [-v]     - verbose" << endl
       << "  [-h]     - print this message" << endl;
}

int main(int argc, char** argv) {
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  string confFileName;
  string resource;
  bool verbose = false;
  int level = 2; 
  int opt;

    while ((opt = getopt(argc,argv,"F:l:r:va")) != EOF) {
      switch (opt) {
        case 'F':
         confFileName = optarg;
         break;
       case 'l':
         level = atoi(optarg);
         break;
       case 'r':
         resource = optarg;
         break;
       case 'v':
         verbose = true;
         break;
       case 'a': 
         level = 3;
         break;
       case 'h':
       default:
         printUsage();
	 exit(2);
      }
    }

    if ((level < 1) || (level > 3)) {
        cerr << "Error: Level not defined" << endl;
        printUsage ();
        exit(2);
    }

    if (!resource.length()) {
	cerr << "no resource name given" << endl;
	printUsage();
	exit(2);
    }

  try {
    if (verbose) {
     cout  << "Initiizing XML library ... " << endl;
    }
 
    XMLUtil::init();
 
    if (verbose) {
      cout  << "Installing security provider ..." << endl;
    }
 
    Provider po = Provider(new OpenSSLProviderImpl);
 
    po->loadSignatureAlgorithms();
    po->seedRandomGenerator("WE GOT TO SEED THE RANDOM GENERATOR");
    Provider::setProvider(po);
 
    if (verbose) {
      cout  << "Loading client configuration file ..." << endl;
    }

    Properties& props = Properties::getSystemProperties();
 
    props.load(confFileName);
 
    if (StringUtil::equalsIgnoreCase(props.getValue("SSLMode"), "on")) {
      if (verbose) {
        cout  << "Setting ssl context ..." << endl;
      }
 
      SSLContext ctx = po->createSSLContext(SSLConstants::TLS1, true);
 
      ctx->setVerify(SSLConstants::VERIFY_PEER);
      ctx->setPrivateKeyFile(props.getValue("ClientKeyFile"), SSLConstants::PEM);
      ctx->setCertificateFile(props.getValue("ClientCertFile"), SSLConstants::PEM);
 
      string path = props.getValue("CertificateAuthorityPath");
 
      if (path.length() != 0) {
        path += "/";
      }
 
      path += props.getValue("CertificateAuthorityFile");
      ctx->setTrustedCAS(path);
      SSLContext::setDefaultSSLClientContext(ctx);
    }

    if (!resource.length()) {
      cerr << "Must provide a resource" << endl;
      exit(2);
    }

    CheckPolicyRequest req;
    CheckPolicyResponse resp;

    req.setID(1);
    req.setResource(resource);
    ProtocolHandler::getInstance(ProtocolHandler::CLIENT).handle(req, resp);

    const vector <PolicyInfo>&  policyInfoVec = resp.getPolicyInfos();

    AkentiMessage mesg = resp.getAkentiMessage();

    bool worked = (mesg.getCode() == 150);

    cout << "\n----------------------" << endl;

    if (worked) {
       cout << "RemoteCheckPolicy WORKED" << endl;
    } else {
       cout << "RemoteCheckPolicy FAILED" << endl;
    }

    cout << "----------------------" << endl;

    if (level == 1) { // we are done
       exit(worked == false);
    }

    for (int j = 0; j < policyInfoVec.size(); ++j) {
        string resource = policyInfoVec[j].getResource();
        const vector <CertificateInfo>& certInfoVec = policyInfoVec[j].getCertificateInfos();

        for (int i = 0; i <certInfoVec.size();  ++i) {
           CertificateInfo certInfo = certInfoVec[i];
           cout << "----------------------" << endl;
    
           string type;
           if (certInfo.hasCertificate()) {
              type = (certInfo.getCertificate().get())->getType();
           }
           else {
              type = "dummy";
           }  
           
           cout << "Resource = " << resource << ", Used = " << certInfo.getUsed(); 
           if (type == "dummy") {
              cout << ", Top level error" << endl;
           } else {
               cout << ", Type = " << type << endl;
           }
    
           if (type == "UseCondition") {
              cout << "Location = " << certInfo.getLocation() << endl;
           }
    
           const vector<AkentiMessage>& mesgs = certInfo.getCertificateMessages();
        
           for (int j = 0; j < mesgs.size(); ++j) {
             cout << mesgs[j].getCode() << ": " << mesgs[j].getMessage() << endl;
           }
    
           if (level == 3) {
               if (type != "dummy") {
                   AkentiCertificate::writeToStream("XML", cout, certInfo.getCertificate());
               }
           }
    
           cout << "----------------------" << endl;
        }
  }

    exit(mesg.getCode() >= 200);
  } catch(Exception& ex) {
    cerr << "Exception:" << ex << endl;
    exit(2);
  } catch(Error& err) {
    cerr << "Error:Report to Akenti team:" << err << endl;
    exit(3);
  } 
}

}
