#include <iostream>
#include <string>
#include <unistd.h>
#include "Provider.hpp"
#include "OpenSSLProviderImpl.hpp"
#include "Certificate.hpp"
#include "CertificateFactory.hpp"
#include "AkentiCertificate.hpp"
#include "File.hpp"
#include "PublicKey.hpp"
#include "Exception.hpp"
#include "Error.hpp"
#include "XMLUtil.hpp"

using namespace std;
using namespace Akenti;

/**
  * <pre>
  * <b>VerifyAkentiCert</b>
  *   Use this program to check the following: 
  *     Check 1: akenti certificate has not expired
  *     Check 2: akenti certificate signature is valid (optional).
  *
  * Note a check is also made on the time validity of the signer's cert
  *      if one is supplied.
  *
  * </pre>
  */

namespace VerifyAkentiCert {

void 
printPurpose() {
  cerr << "Use this program to check the following:" << endl
       << "  Check 1: akenti certificate has not expired" << endl
       << "  Check 2: akenti certificate's signature is valid (optional)" << endl
       << endl
       << "     Note a check is made on the time validity of the signer's cert" << endl
       << "          if one is supplied." << endl
       << endl;
}

void 
printUsage() {
  cerr << "Usage: VerifyAkentiCert args" << endl
       << "  -a arg       - akenti certificate " << endl
       << "  [-i arg]     - issuer's x509 certificate" << endl
       << "  [-v]         - verbose"<< endl 
       << "  [-h]         - prints this message "<< endl; 
}

int
main(int argc, char *argv[]) {
  if (argc < 2) {
    printPurpose();
    printUsage();
    exit(2);
  }

  string akentiCertFileName;
  string x509CertFileName;
  bool verbose = false;
  int opt;

  while ((opt = getopt(argc, argv, "a:i:vh")) != EOF) {
    switch(opt){
      case 'a':
        akentiCertFileName = optarg;
        break;
      case 'i':
        x509CertFileName = optarg;
        break;
      case 'v':
        verbose = true;
        break;
      case 'h':
      default:
        printUsage();
        exit(2);
    }
  }

  try {
    if (verbose) {
      cout << "Initializing XML library ..." << endl;
    }

    XMLUtil::init();

    if (verbose) {
      cout << "Setting security provider ..." << endl;
    }

    Provider po = Provider(new OpenSSLProviderImpl);
    
    po->loadDigestAlgorithms();
    po->loadSignatureAlgorithms();
    Provider::setProvider(po);

    if (!File(akentiCertFileName).isFile()) {
      cerr << "Can't open Akenti cert file" << akentiCertFileName << endl;
      exit(2);
    } 

    if (verbose) {
      cout << "Reading Akenti Certificate " << akentiCertFileName << endl;
    }

    AkentiCertificate cert = AkentiCertificate::initFromFile(akentiCertFileName);

    if (verbose) {
      AkentiCertificate::writeToStream("XML", cout, cert);
    }

    UtcTime now;

    bool ret = cert->checkValidity(now);

    if (!ret) {
      cout << "Akenti Certificate time is not valid as of " << now.toString("YYYY-MM-DDTHH:MM:SSZ") << endl;
    } else {
      cout << "Akenti Certificate time is valid as of " << now.toString("YYYY-MM-DDTHH:MM:SSZ") << endl;
    }

    if (!x509CertFileName.length()) {
      exit(!ret);  
    }

    if (verbose) {
      cout << "Reading issuer's X509 Certificate" << x509CertFileName << endl;
    }

    if (!File(x509CertFileName).isFile()) {
      cerr << "Can't open issuer's X509 cert file" << x509CertFileName << endl; 
      exit(2);
    } 

    CertificateFactory certFac = CertificateFactory::getInstance();
    Certificate x509Cert = certFac->initCertificateFromFile("PEM", x509CertFileName);

    if (verbose) {
      cout << x509Cert->toText() << endl;
    }

    if (!x509Cert->checkValidity(now)) {
      cout << "issuer's certificate is not valid as of " << now << endl;
      exit(1);
    } else if (!cert->verify(x509Cert->getSubjectPublicKey())) {
      cout << "invalid signature" << endl;
      exit(1);
    } else {
      cout << "signature verifies ok" << endl;
      exit(!ret);
    }
  } catch (Exception& ex) {
    cerr << "Exception:" << ex.getMessage() << endl;
    exit(2);
  } catch (Error& err) {
    cerr << "Error:Report to Akenti team:" << err.getMessage() <<endl;
    exit(3);
  }
}

}
