#ifndef AKENTI_ISSUER_HPP
#define AKENTI_ISSUER_HPP

#include <vector>
#include "Directory.hpp"
#include "AkentiPrincipal.hpp"

using std::vector;

namespace Akenti {

class URL;

/**
  * An <i>AkentiIssuer</i> holds an <i>AkentiPrincipal</i>
  * and the directories where its chain of x509 certs is stored.
  *
  * <pre>
  *
  * AkentiIssuer {                    
  *   AkentiPrincipal issuer;         
  *   vectorlt&Directory<&gt  idDirs; 
  * }                                 
  *
  * Format:               
  *   issuer idDirCount idDirs*
  *
  * </pre>
  *
  * @see AkentiObject 
  * @see AkentiPrincipal 
  * @see Directory 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  **/

class AkentiIssuer : public AkentiObject {

public:

  /**
    * Constructs this object.
    *
    **/
  AkentiIssuer(const AkentiPrincipal& issuer  = AkentiPrincipal(),
               const vector<Directory>& idDirs = vector<Directory>());

  /** 
    * Destroys this object.
    **/
  ~AkentiIssuer();

  /**
    * Returns the issuer.
    */
  const AkentiPrincipal& getIssuer() const;
 
  /**
    * Returns the name of the issuer.
    *
    **/
  const DistinguishedName& getName() const;

  /**
    * Returns the name of the issuer's ca.
    **/
  const DistinguishedName& getGuarantor() const;
 
  /**
    * Returns the issuer's <i>x509</i> directories.
    **/
  const vector<Directory>& getIdentityDirs() const;

  /**
    * Returns <i>true</i> if equal.
    *
    * Note: Here we only compare the issuers.
    *
    * @see AkentiPrincipal#operator==()
    **/
  bool operator==(const AkentiIssuer& issuer) const;

  /**
    * Writes this object to an <i>AkentiOutputStream</i>.
    *
    * @see AkentiOutputStream
    **/
  void writeObject(AkentiOutputStream& drain) const;

  /**
    * Reads this object from the <i>AkentiInputStream</i>.
    *
    * @param source AkentiInputStream used to read the object
    *
    * @see AkentiInputStream
    **/
  void readObject(AkentiInputStream& source);

  /**
    * Populates the XML element <i>issuer</i> from this <i>AkentiIssuer</i>.
    */
  void toXML(DOMElement* issuer) const;

  /**
    * Initializes this <i>AkentiIssuer</i> from the XML element
    * <i>issuer</i>.
    */
  void fromXML(const DOMElement* issuer);

  /**
    * Parses the directories and polpulates the <i>urls</i>.
    *
    * @return true if all directories parse
    *         false otherwise
    *
    * @see Directory
    **/
   bool parseIdentityDirs(vector<URL>& urls) const;

  /**
    * Used for debugging.
    */
  string paramString() const;

  /**
    * Used for logging. 
    */
  string toLogString() const;

private:

  AkentiPrincipal issuer;

  vector<Directory> idDirs;

};

} // end of name space

#endif
