#ifndef AK_CERTIFICATE_GENERATOR_IMPL_HPP
#define AK_CERTIFICATE_GENERATOR_IMPL_HPP

#include <string>

using std::string;

namespace Akenti {

class DistinguishedName;
class PublicKey;
class PrivateKey;
class Certificate;

/**
  * This abstract class provides applications with certificate generation.
  * A certificate is a constant object that can not be changed once it has
  * been created. Certficates are either initialized by reading in the contents
  * from a file, or generated by a certificateGenerator object. The generator
  * allows the  various fields of a certificate to be set and when they have 
  * all been set, the generate method is called which creates a new certificate.
  *
  * @see CertificateGenerator
  * @see Certificate
  * @see OpenSSLCertificateGeneratorImpl - a subclass of this one
  * @see AkentiCertGenerator - the base  class for all AkentiCerts
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class CertificateGeneratorImpl {

public:

  /**
    * Constructs this object.
    */
  CertificateGeneratorImpl() {
  }

  /**
    * Deletes this object.
    */
  virtual ~CertificateGeneratorImpl() {
  }
  
  /** 
    * Sets the subject of this certificate.
    * 
    * @param subject DistinguishedName is the  subject of this certificate
    */
  virtual void 
  setSubject(const DistinguishedName& subject) = 0;

  /**
    * Sets the Subject's publicKey.
    *
    * @param subjectKey PublicKey this is the public key of the subject
    */ 
  virtual void 
  setSubjectPublicKey(const PublicKey& subjectKey) = 0;

  /**
    * Sets the issuer of this certificate.
    * 
    * @param issuer the issuer used for setting 
    */ 
  virtual void 
  setIssuer(const DistinguishedName& issuer) = 0;

  /**
    * Sets the validity of this certificate.
    *
    * @param days length of time used for setting this validation
    */
  virtual void 
  setValidity(int days) = 0;

  /**
    * Sets the serial number of this certificate.
    *
    * @param serialNumber long the serial number
    */
  virtual void 
  setSerialNumber(long serialNumber) = 0;

  /**
    * Sets the version number of this certificate.
    *
    * @param versionNumber long the version number
    */ 
  virtual void 
  setVersionNumber(long versionNumber) = 0;

  /**
    * Signs the certificate using the signature algorithm specified 
    * by <i>algorithm</i> and the guarantor's private key specified
    * by <i>privateKey</i>. The signature is added to the certificate fields.
    * Once a certificate has been generated, it  can no longer be modified.
    *
    * @param algorithm string the specified algorithm
    * @param privateKey PrivateKey the private key of the guarantor
    * @return the generated certificate
    * @exception SecurityException 
    */ 
  virtual Certificate 
  generate(const string& algorithm, const PrivateKey& privateKey) = 0;

  /**
    * Returns string of items of this certificate for correctness check.
    */ 
  virtual string paramString() const = 0;

};

} // end of namespace Akenti

#endif
