#ifndef AK_CERTIFICATE_REQUEST_GENERATOR_IMPL_HPP
#define AK_CERTIFICATE_REQUEST_GENERATOR_IMPL_HPP

#include <string>

using std::string;

namespace Akenti {

class DistinguishedName;
class PublicKey;
class PrivateKey;
class CertificateRequest;

/**
  * This abstract class groups all services of setting up 
  * a <i>CertificateRequestGenerator</i> which includes setting subject
  * publicKey, versionNumber, generating CertificateRequest with
  * specified algorithm and privateKey, displaying the Certificate
  * in string. A CertficateRequest is a constant object once it has been 
  * generated and signed. A CertificateRequest is  generally a specification of
  * an X.509 certificate that will be sent  to a CA for it to sign.
  *
  * @see DistinguishedName
  * @see PublicKey
  * @see PrivateKey
  * @see CertificateRequest
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */ 

class CertificateRequestGeneratorImpl {

public:

  /**
    * Construct this object.
    */
  CertificateRequestGeneratorImpl() {
  }

  /**
    * Deletes this object.
    */
  virtual ~CertificateRequestGeneratorImpl() {
  }
  
  /**
    * Sets the subject of the certificate
    *
    * @param subject - the distinguishedName of the principal of the certificate
    */ 
  virtual void 
  setSubject(const DistinguishedName& subject) = 0;

  /**
    * Sets the  publicKey associated with the certificate principal
    *
    * @param publicKey the subject's public key
    */ 
  virtual void 
  setPublicKey(const PublicKey& subjectKey) = 0;

  /**
    * Sets the versionNumber
    *
    * @param versionNumber of the software creating this request
    */
  virtual void 
  setVersionNumber(long versionNumber) = 0;

  /**
    * Generates CertificateRequest using specified algorithm 
    * such as "RSA-SHA1" or "DSA-MD5" and signs the request with the 
    * privateKey which must correspond to the publicKey of the subjet 
    * that is included in the request.
    *
    * @param algorithm hash and signature algorthim to be used in 
    *   signing the request.
    * @param privateKey key of the requestor
    */
  virtual CertificateRequest 
  generate(const string& algorithm, const PrivateKey& privateKey) = 0;

  /**
    * Display the CertificateRequest in string
    */
  virtual string paramString() const = 0;

};

} // end of namespace Akenti

#endif
