#ifndef AK_ERROR_HPP
#define AK_ERROR_HPP

#include <iosfwd>
#include <string>
#include <exception>

using std::string;
using std::exception;
using std::ostream;

/**
  * Errors thrown by the Akenti libraries when it encounters some error
  * that can only be caused by incorrect code in the library. Exceptions
  * are thown for expected run-time errors.
  *
  * @see Exception
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class Error : public exception {

public:

  /**
    * Constructs an <i>Error</i> with the specified 
    * message, a string that best describes this particular 
    * exception.
    *
    * @param message the detail message
    */
  Error(const string &message);

  /**
    * Destroys the <i>Error</i> object.
    */
  virtual ~Error() throw();

  /**
    * Returns the  message as a string. This is preferred 
    * over the what method.
    */
  virtual const string& getMessage() const;

  /**
    * Returns the  message as char*. 
    */
  virtual const char* what() const throw();

  /**
    * Prints the data member for <i>Error</i> using the format
    * [nameOfDataMember="valueOfDataMember",...]. 
    * Double quotes are used to enclose the values of the data members
    * only if they are strings.
    * 
    * @return the string format as shown above
    */
  virtual string paramString() const;
  
private:

  void* operator new(std::size_t);

  string message;    /* describes the exception */

};

} // end namespace Akenti

/**
  * Writes the <i>Error</i> exception message to
  * the stream. Used mainly to print the message to stdout.
  */
ostream& operator<<(ostream& os, const Akenti::Error& exception);


#endif
