#ifndef AK_EXCEPTION_HPP
#define AK_EXCEPTION_HPP

#include <iosfwd>
#include <string>
#include <exception>

using std::string;
using std::exception;
using std::ostream;

/**
  * This is the base class of all Exceptions thrown by the Akenti libraries.
  * Exceptions are thrown when the code discovers errors that may be caused
  * by runtime conditions, such as bad arguments, unreachable network 
  * resources, etc.
  *
  * @see Error
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

namespace Akenti {

class Exception : public exception {

public:

  /**
    * Constructs an <i>Exception</i> with the specified 
    * message, a string that best describes this particular 
    * exception.
    *
    * @param message the detail message
    */
  Exception(const string &message);

  /**
    * Destroys the <i>Exception</i> object.
    */
  virtual ~Exception() throw();

  /**
    * Returns the  message as a string. This is preferred 
    * over the what method.
    */
  virtual const string& getMessage() const;

  /**
    * Returns the  message as char*. 
    */
  virtual const char* what() const throw();

  /**
    * Prints the data member for <i>Exception</i> using the format
    * [nameOfDataMember="valueOfDataMember",...]. 
    * Double quotes are used to enclose the values of the data members
    * only if they are strings.
    * 
    * @return the string format as shown above
    */
  virtual string paramString() const;
  
private:

  void* operator new(std::size_t);

  string message;    /* describes the exception */

};

} // end of namespace Akenti

/**
  * Writes the exception message to
  * the stream. Used mainly to print the message to stdout.
  */
ostream& operator<<(ostream& os, const Akenti::Exception& exception);

#endif
