#ifndef AK_UTCTIME_HPP
#define AK_UTCTIME_HPP

#include <iosfwd>
#include <time.h>
#include <string>

using std::string;
using std::istream;
using std::ostream;

namespace Akenti {

/**
  * A <i>UtcTime</i> object provides the operations for utctime objects.
  * utctime is date and time in Coordinated Universal Time 
  * (aka Greenwich Mean Time). 
  *
  * @version 1.1 00/05/01
  * @author Srilekha Mudumbai
  * @author Abdelilah Essiari
  */

class UtcTime {

public:

  /**
    * Construct this object based on the  string <i>date</i>
    * 
    * @param date const string& the utctime in either YYMMDDHHMMSSZ
    * or YYYY-MM-DDTHH:MM:SS format must end with a "Z"
    */
  UtcTime(const string& date, const string format = string("YYMMDDHHMMSSZ"));

  /**
    * Construct this object based on the <i>offset</i> in seconds from the 
    * Epoch. On UNIX systems the Epoch is 00:00:00 Jan 1, 1970.
    * 
    * @param offset int utctime based on the offset
    */
  UtcTime(int offset = 0);

  /**
    * Destroy this object 
    */
  ~UtcTime();

  /**
    * Converts the utctime to a string 	
    * 
    * @return the utctime in either format
    */
  string toString(const string& format = "YYMMDDHHMMSSZ") const; 

  /**
    * Formats the utctime into a char array
    *
    * @param buffer[] char the char array to which the utctime string was 
    *        written. The format either string format
    */
  void toCharArray(char buffer[], const string format = string("YYMMDDHHMMSSZ")) const; 

  /**
    * Provides an operation to add an offset to the utctime 
    * 
    * @param offset int the specified offset in seconds
    * @return the new utctime 
    */
  UtcTime operator+(int offset) const;

  /**
    * Provides an operation to add an offset to the utctime 
    * 
    * @param offset int the specified offset 
    * @return the new utctime 
    */
  UtcTime& operator+=(int offset);

  /**
    * Checks if utctime object is equal to the input time  
    * 
    * @param utcTime const UtcTime & the utctime to compare with
    * @return true if the utctimes are equal, false if they are not
    */
  bool operator==(const UtcTime &utcTime) const;

  /**
    * Checks If the utcTime object is less than the input time.
    * 
    * @param utcTime const UtcTime & the utctime to compare with
    * @return true if the uctTime in the object is less than the input time
    */
  bool operator<(const UtcTime &utcTime) const;

  /**
    * Checks If the utcTime object is less than or equal to the input
    * time.
    * 
    * @param utcTime const UtcTime & the utctime to compare with
    * @return true if the uctTime in the object is less than or equal
    * to the input time
    */
  bool operator<=(const UtcTime &utcTime) const;

  /**
    * Checks If the utcTime object is greater than the input time.
    * 
    * @param utcTime const UtcTime & the utctime to compare with
    * @return true if the uctTime in the object is greater than the
    * input time
    */
  bool operator>(const UtcTime &utcTime) const;

  /**
    * Checks If the utcTime object is greater than or equal to the
    * input time.
    * 
    * @param utcTime const UtcTime & the utctime to compare with
    * @return true if the uctTime in the object is greater than or
    * equal to the input time
    */
  bool operator>=(const UtcTime &utcTime) const;

  /**
    * Shows the utctime fields in the format 
    * [nameOfDataMember=valueOfDataMember,stringValue].
    * There is no nameOfDataMember for the second value shown above becuase 
    * the other value is just a string representation of the data member.
    * 
    * @return the time_t and the string format of utctime. 
    */
  string paramString() const;

private:

  time_t dtime;

};

} // end of namespace

/**
  * Writes the utcTime to the stream. 
  */ 
ostream& operator<<(ostream& os, const Akenti::UtcTime& utcTime);

/**
  * Reads the utcTime from the stream.
  */
istream& operator>>(istream& is, Akenti::UtcTime& utcTime);

#endif

